from typing import Sequence, List, Dict, Any, Optional

import numpy as np
import pandas as pd
from tabulate import tabulate
import tensorflow as tf

from sknlp.vocab import Vocab
from sknlp.data import TaggingDataset

from ..supervised_model import SupervisedNLPModel
from ..text2vec import Text2vec

from .utils import tagging_fscore


class DeepTagger(SupervisedNLPModel):
    def __init__(
        self,
        classes: Sequence[str],
        start_tag: Optional[str] = None,
        end_tag: Optional[str] = None,
        max_sequence_length: Optional[int] = None,
        sequence_length: Optional[int] = None,
        segmenter: str = "char",
        embedding_size: int = 100,
        text2vec: Optional[Text2vec] = None,
        loss: str = None,
        **kwargs
    ):
        classes = list(classes)
        if start_tag is not None and end_tag is not None:
            self._start_tag = start_tag
            self._end_tag = end_tag
            if start_tag not in classes:
                classes.append(start_tag)
            if end_tag not in classes:
                classes.append(end_tag)
        self._pad_tag = "[PAD]"
        if self._pad_tag != classes[0]:
            classes.insert(0, self._pad_tag)
        super().__init__(
            classes,
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=segmenter,
            embedding_size=embedding_size,
            text2vec=text2vec,
            task="tagging",
            **kwargs
        )
        self._loss = loss

    def get_loss(self) -> None:
        return None

    def get_callbacks(self) -> List[tf.keras.callbacks.Callback]:
        return []

    def get_metrics(self) -> List[tf.keras.metrics.Metric]:
        return []

    def get_monitor(self) -> str:
        return None

    def create_dataset_from_df(
        self, df: pd.DataFrame, vocab: Vocab, segmenter: str, labels: Sequence[str]
    ) -> TaggingDataset:
        return TaggingDataset(
            vocab,
            list(labels),
            df=df,
            max_length=self._max_sequence_length,
            text_segmenter=segmenter,
            start_tag=self._start_tag,
            end_tag=self._end_tag,
        )

    def dummy_y(self, X: Sequence[str]) -> List[List[str]]:
        return [["O" for _ in range(len(xi))] for xi in X]

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: TaggingDataset = None,
        batch_size: int = 128
    ) -> List[List[str]]:
        tag_ids_list = super().predict(X, dataset=dataset, batch_size=batch_size)
        exclude_tags = {self._pad_tag, self._start_tag, self._end_tag}
        predictions = []
        for tag_ids in tag_ids_list:
            tags = []
            for tag_id in tag_ids:
                tag = self._idx2class[tag_id]
                if tag not in exclude_tags:
                    tags.append(tag)
                if tag == self._pad_tag:
                    break
            predictions.append(tags)
        return predictions

    def score(
        self,
        X: Sequence[str] = None,
        y: Sequence[Sequence[str]] = None,
        *,
        dataset: TaggingDataset = None,
        batch_size: int = 128
    ) -> pd.DataFrame:
        predictions = self.predict(X=X, dataset=dataset, batch_size=batch_size)
        labels = [c.split("-")[-1] for c in self.classes if "-" in c]
        return tagging_fscore(X, y, predictions, labels)

    def format_score(self, score_df: pd.DataFrame, format: str = "markdown") -> str:
        return tabulate(score_df, headers="keys", tablefmt="github", showindex=False)

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "start_tag": self._start_tag,
            "end_tag": self._end_tag,
        }

    @classmethod
    def _filter_config(cls, config):
        config = super()._filter_config(config)
        config.pop("algorithm", None)
        config.pop("task", None)
        return config

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
        }


from typing import Tuple, List, Sequence
from dataclasses import dataclass
import itertools

import numpy as np
import pandas as pd


@dataclass
class Tag:
    label: str
    start: int
    end: str
    text: str

    def __hash__(self):
        return hash((self.label, self.start, self.end, self.text))


def parse_tagged_text(text: str, tag_names: Sequence[str]) -> List[Tag]:
    current_label = None
    start = 0
    parsed_tags = list()
    for i, tag_name in enumerate(tag_names):
        if tag_name == "O":
            tag_type, tag_label = "O", None
        else:
            tag_type, tag_label = tag_name.split("-")
        if (tag_type == "I" or tag_type == "E") and tag_label == current_label:
            continue

        if i != start:
            parsed_tags.append(Tag(current_label, start, i, text[start:i]))

        if tag_type == "B" or tag_type == "S":
            start = i
            current_label = tag_label
        else:
            start = i + 1
            current_label = None
    if start != len(tag_names) and current_label is not None:
        parsed_tags.append(Tag(current_label, start, len(tag_names), text[start:]))
    return parsed_tags


def _compute_counts(
    text: str, label: Sequence[str], prediction: Sequence[str], classes: Sequence[str]
) -> List[Tuple[str, int, int, int]]:
    truth_tags = set(parse_tagged_text(text, label))
    prediction_tags = set(parse_tagged_text(text, prediction))
    correct_tags = truth_tags & prediction_tags
    counts = [
        (
            cls,
            len([i for i, tag in enumerate(correct_tags) if tag.label == cls]),
            len([i for i, tag in enumerate(truth_tags) if tag.label == cls]),
            len([i for i, tag in enumerate(prediction_tags) if tag.label == cls]),
        )
        for cls in classes
    ]
    counts.append(("avg", len(correct_tags), len(truth_tags), len(prediction_tags)))
    return counts


def tagging_fscore(
    texts: Sequence[str],
    y: Sequence[Sequence[str]],
    p: Sequence[Sequence[str]],
    classes: Sequence[str],
) -> pd.DataFrame:
    df = pd.DataFrame(
        itertools.chain.from_iterable(
            _compute_counts(text, yi, pi, classes) for text, yi, pi in zip(texts, y, p)
        ),
        columns=["class", "correct", "support", "prediction"],
    )
    df = df.groupby("class").sum()
    df["precision"] = df.correct / df.prediction
    df["recall"] = df.correct / df.support
    df["fscore"] = 2 * df.precision * df.recall / (df.precision + df.recall)
    df.fillna(0, inplace=True)
    df.reset_index(inplace=True)
    return df[["class", "precision", "recall", "fscore", "support"]]



from typing import Any, Dict, Sequence, Optional

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Dropout
import pandas as pd

from sknlp.layers import (
    MLPLayer,
    BertLayer,
    BertCharPreprocessingLayer,
    CrfLossLayer,
    CrfDecodeLayer,
)
from sknlp.vocab import Vocab
from sknlp.data import BertTaggingDataset
from ..text2vec import Bert2vec
from .deep_tagger import DeepTagger


class BertTagger(DeepTagger):
    def __init__(
        self,
        classes: Sequence[str],
        segmenter: Optional[str] = None,
        embedding_size: int = 100,
        max_sequence_length: int = 120,
        num_fc_layers: int = 2,
        fc_hidden_size: int = 256,
        output_dropout: float = 0.5,
        text2vec: Optional[Bert2vec] = None,
        **kwargs
    ) -> None:
        super().__init__(
            classes,
            segmenter=segmenter,
            algorithm="bert",
            embedding_size=embedding_size,
            max_sequence_length=max_sequence_length,
            start_tag="[CLS]",
            end_tag="[SEP]",
            text2vec=text2vec,
            **kwargs
        )
        if self._text2vec is not None:
            self.preprocessing_layer = BertCharPreprocessingLayer(
                self._text2vec.vocab.sorted_tokens
            )
        self.mlp_layer = MLPLayer(
            num_fc_layers,
            hidden_size=fc_hidden_size,
            output_size=self.num_classes,
            name="mlp",
        )
        self.crf_layer = CrfLossLayer(self.num_classes)
        self.inputs = [
            tf.keras.Input(shape=(), dtype=tf.string, name="text_input"),
            tf.keras.Input(shape=(None,), dtype=tf.int32, name="tag_id"),
        ]
        self._output_dropout = output_dropout

    def create_dataset_from_df(
        self, df: pd.DataFrame, vocab: Vocab, segmenter: str, labels: Sequence[str]
    ) -> BertTaggingDataset:
        return BertTaggingDataset(
            vocab,
            list(labels),
            df=df,
            max_length=self._max_sequence_length,
        )

    def get_inputs(self) -> tf.Tensor:
        return self.inputs

    def get_outputs(self) -> tf.Tensor:
        return self.build_output_layer(self.build_encode_layer(self.get_inputs()))

    def build_encode_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        texts, tag_ids = inputs
        token_ids = self.preprocessing_layer(texts)
        embeddings, mask, _ = self._text2vec(
            [token_ids, K.zeros_like(token_ids, dtype=tf.int64)]
        )
        mask = tf.keras.layers.Lambda(
            lambda x: tf.cast(x, tf.int32), name="mask_layer"
        )(mask)
        if self._output_dropout:
            noise_shape = (None, 1, self.embedding_size)
            embeddings = Dropout(
                self._output_dropout,
                noise_shape=noise_shape,
                name="embedding_dropout",
            )(embeddings)
        return embeddings, mask, tag_ids

    def build_output_layer(self, inputs: tf.Tensor) -> tf.Tensor:
        embeddings, mask, tag_ids = inputs
        emissions = self.mlp_layer(embeddings)
        return self.crf_layer([emissions, tag_ids], mask)

    @classmethod
    def _filter_config(cls, config: Dict[str, Any]) -> Dict[str, Any]:
        config = super()._filter_config(config)
        config.pop("start_tag", None)
        config.pop("end_tag", None)
        return config

    def export(self, directory: str, name: str, version: str = "0") -> None:
        mask = self._model.get_layer("mask_layer").output
        emissions = self._model.get_layer("mlp").output
        crf = CrfDecodeLayer(self.num_classes, self._max_sequence_length)
        crf.set_weights(self._model.get_layer("crf").get_weights())
        model = tf.keras.Model(
            inputs=self._model.inputs[0], outputs=crf(emissions, mask)
        )
        original_model = self._model
        self._model = model
        super().export(directory, name, version=version)
        self._model = original_model

    def get_config(self) -> Dict[str, Any]:
        return {**super().get_config(), "output_dropout": self._output_dropout}

    def get_custom_objects(self) -> Dict[str, Any]:
        return {
            **super().get_custom_objects(),
            "MLPLayer": MLPLayer,
            "BertLayer": BertLayer,
            "BertCharPreprocessingLayer": BertCharPreprocessingLayer,
            "CrfLossLayer": CrfLossLayer,
            "CrfDecodeLayer": CrfDecodeLayer,
        }



from typing import Sequence, Union, Optional, Dict, Any, List

import os
import logging

import tensorflow as tf
from tensorflow.python.keras.engine import training
import tensorflow_addons as tfa
import pandas as pd

from ..data import NLPDataset
from ..data.text_segmenter import get_segmenter
from ..vocab import Vocab
from ..callbacks import WeightDecayScheduler
from .text2vec import Word2vec, Text2vec
from .base_model import BaseNLPModel

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream = logging.StreamHandler()
logger.addHandler(stream)


class SupervisedNLPModel(BaseNLPModel):
    def __init__(
        self,
        classes: Sequence[str],
        max_sequence_length: int = None,
        sequence_length: Optional[int] = None,
        segmenter: str = "jieba",
        embedding_size: int = 100,
        text2vec: Optional[Text2vec] = None,
        task: Optional[str] = None,
        algorithm: Optional[str] = None,
        **kwargs
    ) -> None:
        super().__init__(
            max_sequence_length=max_sequence_length,
            sequence_length=sequence_length,
            segmenter=text2vec.segmenter if text2vec else segmenter,
        )
        self._text2vec = text2vec
        if text2vec is not None:
            self._embedding_size = text2vec.embedding_size
            self._max_sequence_length = (
                max_sequence_length or text2vec.max_sequence_length
            )
            self._sequence_length = text2vec.sequence_length
        else:
            self._embedding_size = embedding_size
            self._max_sequence_length = max_sequence_length
            self._sequence_length = sequence_length

        self._task = task
        self._algorithm = algorithm
        self._class2idx = dict(zip(list(classes), range(len(classes))))
        self._idx2class = dict(zip(range(len(classes)), list(classes)))

    @property
    def embedding_size(self) -> int:
        return self._embedding_size

    @property
    def num_classes(self) -> int:
        return len(self._class2idx)

    @property
    def classes(self) -> List[str]:
        return list(self._class2idx.keys())

    def get_inputs(self) -> tf.Tensor:
        return self._text2vec.get_inputs()

    def get_outputs(self) -> tf.Tensor:
        inputs = self._text2vec.get_outputs()
        return self.build_output_layer(self.build_encode_layer(inputs))

    def build(self) -> None:
        if self._built:
            return
        assert self._text2vec is not None
        super().build()

    @classmethod
    def create_dataset_from_df(
        cls, df: pd.DataFrame, vocab: Vocab, segmenter: str, labels: Sequence[str]
    ) -> NLPDataset:
        raise NotImplementedError()

    def prepare_tf_dataset(
        self,
        X: Sequence[str],
        y: Sequence[str],
        dataset: NLPDataset,
        batch_size: int,
        shuffle: bool = True,
    ) -> tf.data.Dataset:
        assert not ((X is None or y is None) and dataset is None)

        if self._text2vec is None:
            cut = get_segmenter(self._segmenter)
            vocab = self.build_vocab(X, cut)
            self._text2vec = Word2vec(vocab, self._embedding_size, self._segmenter)

        if dataset is not None:
            return dataset.batchify(batch_size, shuffle=shuffle)

        if isinstance(y[0], (list, tuple)):
            y = ["|".join(map(str, y_i)) for y_i in y]
        df = pd.DataFrame(zip(X, y), columns=["text", "label"])
        dataset = self.create_dataset_from_df(
            df, self._text2vec.vocab, self._text2vec.segmenter, self.classes
        )
        return dataset.batchify(
            batch_size, shuffle=shuffle, shuffle_buffer_size=df.shape[0]
        )

    def fit(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        valid_X: Sequence[str] = None,
        valid_y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        valid_dataset: NLPDataset = None,
        batch_size: int = 128,
        n_epochs: int = 30,
        optimizer: str = "adam",
        lr: float = 1e-3,
        weight_decay: float = 1e-4,
        lr_update_factor: float = 0.5,
        lr_update_epochs: int = 10,
        clip: float = 5.0,
        enable_early_stopping: bool = False,
        early_stopping_patience: Optional[int] = None,
        early_stopping_min_delta: float = 0,
        early_stopping_use_best_epoch: bool = False,
        early_stopping_monitor: int = 2,  # 1 for loss, 2 for metric
        checkpoint: Optional[str] = None,
        save_frequency: int = 1,
        log_file: Optional[str] = None,
        verbose: int = 2
    ) -> None:
        assert not (
            self._text2vec is None and X is None
        ), "When text2vec is not given, X must be provided"

        self.train_dataset = self.prepare_tf_dataset(
            X,
            y,
            dataset,
            batch_size,
        )
        if (valid_X is None or valid_y is None) and valid_dataset is None:
            self.valid_dataset = None
        else:
            self.valid_dataset = self.prepare_tf_dataset(
                valid_X,
                valid_y,
                valid_dataset,
                batch_size,
                shuffle=False,
            )
        if not self._built:
            self.build()
        optimizer = tfa.optimizers.AdamW(weight_decay, learning_rate=lr, clipnorm=clip)
        if optimizer != "adam":
            pass
        loss = self.get_loss()
        metrics = self.get_metrics()
        callbacks = self.get_callbacks()
        self._model.compile(optimizer=optimizer, loss=loss, metrics=metrics)

        def lr_scheduler(epoch, lr):
            return lr * lr_update_factor ** ((epoch + 1) % lr_update_epochs == 0)

        def wd_scheduler(epoch, wd):
            return wd * lr_update_factor ** ((epoch + 1) % lr_update_epochs == 0)

        lr_decay = tf.keras.callbacks.LearningRateScheduler(
            lr_scheduler, verbose=verbose
        )
        weight_decay = WeightDecayScheduler(wd_scheduler, verbose=verbose)
        callbacks = [*callbacks, lr_decay, weight_decay]
        if enable_early_stopping:
            if early_stopping_monitor == 2 and self.get_monitor():
                mode = "max"
                monitor = self.get_monitor()
            else:
                mode = "min"
                monitor = "val_loss"
            callbacks.append(
                tf.keras.callbacks.EarlyStopping(
                    monitor=monitor,
                    min_delta=early_stopping_min_delta,
                    patience=early_stopping_patience or lr_update_epochs,
                    mode=mode,
                    restore_best_weights=early_stopping_use_best_epoch,
                )
            )
        if log_file is not None:
            callbacks.append(tf.keras.callbacks.CSVLogger(log_file))
        self._model.fit(
            self.train_dataset,
            epochs=n_epochs,
            validation_data=self.valid_dataset,
            callbacks=callbacks,
            verbose=verbose,
        )

    def dummy_y(self, X: Sequence[str]) -> List[Any]:
        raise NotImplementedError()

    def predict(
        self,
        X: Sequence[str] = None,
        *,
        dataset: NLPDataset = None,
        batch_size: int = 128
    ) -> tf.Tensor:
        assert self._built
        y = None if X is None else self.dummy_y(X)
        dataset = self.prepare_tf_dataset(X, y, dataset, batch_size, shuffle=False)
        return self._model.predict(dataset)

    def score(
        self,
        X: Sequence[str] = None,
        y: Union[Sequence[Sequence[str]], Sequence[str]] = None,
        *,
        dataset: NLPDataset = None,
        thresholds: Union[float, Dict[str, float]] = 0.5,
        batch_size: int = 128
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def get_config(self) -> Dict[str, Any]:
        return {
            **super().get_config(),
            "classes": self.classes,
            "task": self._task,
            "algorithm": self._algorithm,
        }

    def save(self, directory: str) -> None:
        super().save(directory)
        self._text2vec.save_vocab(directory)

    @classmethod
    def load(cls, directory: str) -> "SupervisedNLPModel":
        module = super().load(directory)
        with open(os.path.join(directory, "vocab.json")) as f:
            vocab = Vocab.from_json(f.read())
        module._text2vec = Text2vec(
            vocab,
            segmenter=module.segmenter,
            max_sequence_length=module.max_sequence_length,
            sequence_length=module.sequence_length,
        )
        return module

    def export(self, directory: str, name: str, version: str = "0") -> None:
        super().export(directory, name, version)
        d = os.path.join(directory, name, version)
        self._text2vec.save_vocab(d)

    def get_custom_objects(self) -> Dict[str, Any]:
        return {**super().get_custom_objects(), "AdamW": tfa.optimizers.AdamW}
